@@ -15,7 +15,7 @@
 Module on physical information of the system.
 """
 
-from typing import Callable, List, Optional, Text
+from typing import Callable, List, Optional, Text, Tuple
 from copy import copy
 from functools import lru_cache
 
@@ -26,32 +26,6 @@ from oqupy.base_api import BaseAPIClass
 from oqupy.config import NpDtype
 import oqupy.operators as opr
 
-def _check_hamiltonian(hamiltonian):
-    """Input checking for a single Hamiltonian. """
-    try:
-        tmp_hamiltonian = np.array(hamiltonian, dtype=NpDtype)
-        tmp_hamiltonian.setflags(write=False)
-    except Exception as e:
-        raise AssertionError("Coupling operator must be numpy array") from e
-    assert len(tmp_hamiltonian.shape) == 2, \
-        "Coupling operator is not a matrix."
-    assert tmp_hamiltonian.shape[0] == \
-        tmp_hamiltonian.shape[1], \
-        "Coupling operator is not a square matrix."
-    return tmp_hamiltonian
-
-
-def _liouvillian(hamiltonian, gammas, lindblad_operators):
-    """Lindbladian for a specific Hamiltonian, gammas and lindblad_operators.
-    """
-    liouvillian = -1j * opr.commutator(hamiltonian)
-    for gamma, op in zip(gammas, lindblad_operators):
-        op_dagger = op.conjugate().T
-        liouvillian += gamma * (opr.left_right_super(op, op_dagger) \
-                                - 0.5 * opr.acommutator(np.dot(op_dagger, op)))
-    return liouvillian
-
-
 class BaseSystem(BaseAPIClass):
     """Base class for systems. """
     def __init__(
@@ -68,35 +42,6 @@ class BaseSystem(BaseAPIClass):
         """Hilbert space dimension of the system. """
         return self._dimension
 
-    def liouvillian(self, t: Optional[float] = None) -> ndarray:
-        r"""
-        Returns the Liouvillian super-operator :math:`\mathcal{L}(t)` with
-
-        .. math::
-
-            \mathcal{L}(t)\rho = -i [\hat{H}(t), \rho]
-                + \sum_n^N \gamma_n \left(
-                    \hat{A}_n(t) \rho \hat{A}_n^\dagger(t)
-                    - \frac{1}{2} \hat{A}_n^\dagger(t) \hat{A}_n(t) \rho
-                    - \frac{1}{2} \rho \hat{A}_n^\dagger(t) \hat{A}_n(t)
-                  \right),
-
-        with time :math:`t`.
-
-        Parameters
-        ----------
-        t: float (default = None)
-            time :math:`t`.
-
-        Returns
-        -------
-        liouvillian : ndarray
-            Liouvillian :math:`\mathcal{L}(t)` at time :math:`t`.
-        """
-        raise NotImplementedError(
-            "Class {} has no liouvillian implementation.".format(
-                type(self).__name__))
-
 class System(BaseSystem):
     r"""
     Represents a system (without any coupling to a non-Markovian bath).
@@ -128,6 +73,7 @@ class System(BaseSystem):
     description: str
         An optional description of the system.
     """
+
     def __init__(
             self,
             hamiltonian: ndarray,
@@ -139,41 +85,14 @@ class System(BaseSystem):
         # input check for Hamiltonian.
         self._hamiltonian = _check_hamiltonian(hamiltonian)
         tmp_dimension = self._hamiltonian.shape[0]
-
         # input check gammas and lindblad_operators
-        if gammas is None:
-            gammas = []
-        if lindblad_operators is None:
-            lindblad_operators = []
-        assert isinstance(gammas, list), \
-            "Argument `gammas` must be a list)]."
-        assert isinstance(lindblad_operators, list), \
-            "Argument `lindblad_operators` must be a list."
-        assert len(gammas) == len(lindblad_operators), \
-            "Lists `gammas` and `lindblad_operators` must have the same length."
-        try:
-            tmp_gammas = []
-            for gamma in gammas:
-                tmp_gammas.append(float(gamma))
-        except Exception as e:
-            raise AssertionError("All elements of `gammas` must be floats.") \
-                from e
-        try:
-            tmp_lindblad_operators = []
-            for lindblad_operator in lindblad_operators:
-                tmp_lindblad_operators.append(
-                    np.array(lindblad_operator, dtype=NpDtype))
-        except Exception as e:
-            raise AssertionError(
-                "All elements of `lindblad_operators` must be numpy arrays.") \
-                    from e
-        self._gammas = tmp_gammas
-        self._lindblad_operators = tmp_lindblad_operators
-
+        self._gammas, self._lindblad_operators = \
+                _check_gammas_lindblad_operators(gammas,
+                        lindblad_operators)
         super().__init__(tmp_dimension, name, description)
 
     @lru_cache(4)
-    def liouvillian(self, t: Optional[float] = None) -> ndarray:
+    def liouvillian(self) -> ndarray:
         r"""
         Returns the Liouvillian super-operator :math:`\mathcal{L}` with
 
@@ -210,7 +129,6 @@ class System(BaseSystem):
         """List of lindblad operators."""
         return copy(self._lindblad_operators)
 
-
 class TimeDependentSystem(BaseSystem):
     r"""
     Represents an explicitly time dependent system (without any coupling to a
@@ -253,55 +171,19 @@ class TimeDependentSystem(BaseSystem):
                 Optional[List[Callable[[float], ndarray]]] = None,
             name: Optional[Text] = None,
             description: Optional[Text] = None) -> None:
-        """Create a System object."""
+        """Create a TimeDependentSystem object."""
         # input check for Hamiltonian.
-        try:
-            tmp_hamiltonian = np.vectorize(hamiltonian)
-            _check_hamiltonian(tmp_hamiltonian(1.0))
-        except Exception as e:
-            raise AssertionError(
-                "Time dependent Hamiltonian must be vectorizable callable.") \
-                    from e
-        self._hamiltonian = tmp_hamiltonian
+        self._hamiltonian = _check_tdependent_hamiltonian(hamiltonian)
         tmp_dimension = self._hamiltonian(1.0).shape[0]
-
         # input check gammas and lindblad_operators
-        if gammas is None:
-            gammas = []
-        if lindblad_operators is None:
-            lindblad_operators = []
-        assert isinstance(gammas, list), \
-            "Argument `gammas` must be a list)]."
-        assert isinstance(lindblad_operators, list), \
-            "Argument `lindblad_operators` must be a list."
-        assert len(gammas) == len(lindblad_operators), \
-            "Lists `gammas` and `lindblad_operators` must have the same length."
-        try:
-            tmp_gammas = []
-            for gamma in gammas:
-                float(gamma(1.0))
-                tmp_gamma = np.vectorize(gamma)
-                tmp_gammas.append(tmp_gamma)
-        except Exception as e:
-            raise AssertionError(
-                "All elements of `gammas` must be vectorizable " \
-                 + "callables returning floats.") from e
-        try:
-            tmp_lindblad_operators = []
-            for lindblad_operator in lindblad_operators:
-                tmp_lindblad_operator = np.vectorize(lindblad_operator)
-                np.array(tmp_lindblad_operator(1.0))
-                tmp_lindblad_operators.append(tmp_lindblad_operator)
-        except Exception as e:
-            raise AssertionError(
-                "All elements of `lindblad_operators` must be vectorizable " \
-                + "callables returning numpy arrays.") from e
-        self._gammas = tmp_gammas
-        self._lindblad_operators = tmp_lindblad_operators
+        self._gammas, self._lindblad_operators = \
+            _check_tdependent_gammas_lindblad_operators(
+                    gammas,
+                    lindblad_operators)
 
         super().__init__(tmp_dimension, name, description)
 
-    def liouvillian(self, t: Optional[float] = None) -> ndarray:
+    def liouvillian(self, t: float) -> ndarray:
         r"""
         Returns the Liouvillian super-operator :math:`\mathcal{L}(t)` with
 
@@ -325,15 +207,7 @@ class TimeDependentSystem(BaseSystem):
         -------
         liouvillian : ndarray
             Liouvillian :math:`\mathcal{L}(t)` at time :math:`t`.
-
-        Raises
-        ------
-        ValueError
-            If `t = None`
         """
-        if t is None:
-            raise ValueError("Liouvillian depends on time: Argument `t` "
-                             + "must be float.")
         hamiltonian = self._hamiltonian(t)
         gammas = [gamma(t) for gamma in self._gammas]
         lindblad_operators = [l_op(t) for l_op in self._lindblad_operators]
@@ -354,6 +228,173 @@ class TimeDependentSystem(BaseSystem):
         """List of lindblad operators. """
         return copy(self._lindblad_operators)
 
+class TimeDependentSystemWithField(BaseSystem):
+    r"""
+    Represents an explicitly time dependent system (without any coupling
+    to a non-Markovian bath) with a coherent field (a complex scalar)
+    :math:`\langle a \rangle` that evolves according to a specified
+    equation of motion :math:`\partial_t\langle a \rangle`.
+
+    It is possible to include (also explicitly time dependent) Lindblad
+    terms in the master equation.  The equations of motion for a system
+    density matrix (without any coupling to a non-Markovian bath) is
+    then:
+
+    .. math::
+
+        \frac{d}{dt}\rho(t) = &-i [\hat{H}(t, \langle a \rangle), \rho(t)] \\
+            &+ \sum_n^N \gamma_n(t) \left(
+                \hat{A}_n(t) \rho(t) \hat{A}_n(t)^\dagger
+                - \frac{1}{2} \hat{A}_n^\dagger(t) \hat{A}_n(t) \rho(t)
+                - \frac{1}{2} \rho(t) \hat{A}_n^\dagger(t) \hat{A}_n(t) \right)
+
+    with the  `hamiltionian` :math:`\hat{H}(t, \langle a \rangle)`
+    depending on both time :math:`t` and `field` :math:`\langle
+    a \rangle` (in general), the  time dependent rates `gammas`
+    :math:`\gamma_n(t)` and the time dependent `linblad_operators`
+    :math:`\hat{A}_n(t)`.
+
+    Parameters
+    ----------
+    hamiltonian: callable
+        System-only Hamiltonian :math:`\hat{H}(t, \langle a \rangle)`
+        where :math:`\langle a \rangle` is the field at time :math:`t`.
+    field_eom: callable
+        Field equation of motion :math:`\partial_t
+        \langle a \rangle(t, \rho, \langle a \rangle)`
+        where :math:`\rho` and :math:`\langle a \rangle` are the system
+        density matrix (a square matrix) and field at time :math:`t`.
+    gammas: list(callable)
+        The rates :math:`\gamma_n(t)`.
+    lindblad_operators: list(callable)
+        The Lindblad operators :math:`\hat{A}_n(t)`.
+    name: str
+        An optional name for the system.
+    description: str
+        An optional description of the system.
+    """
+    def __init__(
+            self,
+            hamiltonian: Callable[[float, complex], ndarray],
+            field_eom: Callable[[float, ndarray, complex], complex],
+            gammas: \
+                Optional[List[Callable[[float], float]]] = None,
+            lindblad_operators: \
+                Optional[List[Callable[[float], ndarray]]] = None,
+            name: Optional[Text] = None,
+            description: Optional[Text] = None) -> None:
+        """Create a TimeDependentSystemWithField object."""
+        # input check for Hamiltonian
+        self._hamiltonian = _check_tfielddependent_hamiltonian(hamiltonian)
+        tmp_dimension = self._hamiltonian(1.0, 1.0+1.0j).shape[0]
+        # input check for field equation of motion
+        tmp_field_eom = _check_field_eom(tmp_dimension, field_eom)
+        self._field_eom = tmp_field_eom
+        # input check gammas and lindblad_operators
+        self._gammas, self._lindblad_operators = \
+             _check_tdependent_gammas_lindblad_operators(
+                     gammas,
+                     lindblad_operators)
+
+        super().__init__(tmp_dimension, name, description)
+
+    def _linearised_hamiltonian(self, t0: float, t: float, state: ndarray,
+            field: complex) -> complex:
+        r"""
+        Return value of the system Hamiltonian at time `t` using a linearisation
+        of the field from its value at time `t0`.
+        """
+        return self._hamiltonian(t, self._linearised_field(t0, t, state, field))
+
+    def _linearised_field(self, t0: float, t: float, state: ndarray,
+            field: complex) -> complex:
+        r"""
+        Return the value of the field a time `(t-t0)` given the value at `t0`
+        in a linear approximation using the field equation of motion.
+        """
+        # field_eom expects a square matrix
+        state = state.reshape((self._dimension, self._dimension))
+        return field + self._field_eom(t0, state, field) * (t-t0)
+
+    def liouvillian(self,
+            t0: float,
+            t: float,
+            state: ndarray,
+            field: complex) -> ndarray:
+        r"""
+        Returns the Liouvillian super-operator :math:`\mathcal{L}(t)` such that
+
+        .. math::
+
+            \mathcal{L}(t)\rho = -i [\hat{H}(t, \langle a \rangle), \rho]
+                + \sum_n^N \gamma_n \left(
+                    \hat{A}_n(t) \rho \hat{A}_n^\dagger(t)
+                    - \frac{1}{2} \hat{A}_n^\dagger(t) \hat{A}_n(t) \rho
+                    - \frac{1}{2} \rho \hat{A}_n^\dagger(t) \hat{A}_n(t)
+                  \right),
+
+        with time :math:`t`.
+
+        Parameters
+        ----------
+        t0: float
+            Start time of the current step
+        t: float
+            Current time :math:`t`
+        state: ndarray
+            Flattened system density matrix (vector) at time :math:`t`
+        field: complex
+            Field value at time :math:`t` obtained from the
+            linearisation of the field at :math:`t` using the field
+            equation of motion.
+
+        Returns
+        -------
+        liouvillian : ndarray
+            Liouvillian :math:`\mathcal{L}(t)` at time :math:`t`.
+        """
+        try:
+            t0 = float(t0)
+        except Exception as e:
+            raise AssertionError("Argument t0 must be float") from e
+        try:
+            t = float(t)
+        except Exception as e:
+            raise AssertionError("Argument t must be float") from e
+        assert t >= t0, "Argument t must exceed t0"
+        try:
+            field = complex(field)
+        except Exception as e:
+            raise AssertionError("Argument field must be complex") from e
+        assert isinstance(state, ndarray), "Argument state must be ndarray"
+        assert state.size ==  self._dimension**2, "State size must match "\
+                "Hilbert space dimension {}".format(self._dimension**2)
+        # flatten if given as a square matrix
+        state = state.reshape((self._dimension**2,))
+        hamiltonian = self._linearised_hamiltonian(t0, t, state, field)
+        gammas = [gamma(t) for gamma in self._gammas]
+        lindblad_operators = [l_op(t) for l_op in self._lindblad_operators]
+        return _liouvillian(hamiltonian, gammas, lindblad_operators)
+
+    @property
+    def hamiltonian(self) -> Callable[[float, complex], ndarray]:
+        """The system Hamiltonian. """
+        return copy(self._hamiltonian)
+
+    @property
+    def gammas(self) -> List[Callable[[float], float]]:
+        """List of gammas. """
+        return copy(self._gammas)
+
+    @property
+    def field_eom(self) -> Callable[[float, float, ndarray, complex], complex]:
+        """The field equation of motion. """
+        return copy(self._field_eom)
+
+    @property
+    def lindblad_operators(self) -> List[Callable[[float], ndarray]]:
+        """List of lindblad operators. """
+        return copy(self._lindblad_operators)
 
 class SystemChain(BaseAPIClass):
     """
@@ -388,7 +429,8 @@ class SystemChain(BaseAPIClass):
         self._nn_liouvillians = []
         for hs_dim_l, hs_dim_r in zip(self._hs_dims[:-1], self._hs_dims[1:]):
             self._nn_liouvillians.append(
-                np.zeros((hs_dim_l**4, hs_dim_r**4), dtype=NpDtype))
+                np.zeros((hs_dim_l**2 * hs_dim_r**2, hs_dim_l**2 * hs_dim_r**2),
+                dtype=NpDtype))
 
         super().__init__(name, description)
 
@@ -455,7 +497,7 @@ class SystemChain(BaseAPIClass):
         liouvillian: ndarray
             Liouvillian acting on the single site.
         """
-        raise NotImplementedError()
+        self._site_liouvillians[site] += np.array(liouvillian, dtype=NpDtype)
 
     def add_site_dissipation(
             self,
@@ -484,12 +526,13 @@ class SystemChain(BaseAPIClass):
         gamma: float
             Optional multiplicative factor :math:`\gamma`.
         """
-        op = lindblad_operator
+        op = np.array(lindblad_operator, dtype=NpDtype)
         op_dagger = op.conjugate().T
         self._site_liouvillians[site] += \
-            gamma * (opr.left_right_super(op, op_dagger)
+            gamma * (opr.left_right_super(op, op_dagger) \
                       - 0.5 * opr.acommutator(np.dot(op_dagger, op)))
 
+
     def add_nn_hamiltonian(
             self,
             site: int,
@@ -544,7 +587,7 @@ class SystemChain(BaseAPIClass):
         liouvillian_l_r: ndarray
             Liouvillian acting on sites :math:`n` and :math:`n+1`.
         """
-        self._nn_liouvillians[site] += liouvillian_l_r
+        self._nn_liouvillians[site] += np.array(liouvillian_l_r, dtype=NpDtype)
 
     def add_nn_dissipation(
             self,
@@ -629,3 +672,142 @@ class SystemChain(BaseAPIClass):
             nn_full_liouvillians.append(nn_full_liouvillian)
 
         return nn_full_liouvillians
+
+def _check_hamiltonian(hamiltonian) -> ndarray:
+    """Input checking for a single Hamiltonian. """
+    try:
+        tmp_hamiltonian = np.array(hamiltonian, dtype=NpDtype)
+        tmp_hamiltonian.setflags(write=False)
+    except Exception as e:
+        raise AssertionError("Coupling operator must be numpy array") from e
+    assert len(tmp_hamiltonian.shape) == 2, \
+        "Coupling operator is not a matrix."
+    assert tmp_hamiltonian.shape[0] == \
+        tmp_hamiltonian.shape[1], \
+        "Coupling operator is not a square matrix."
+    return tmp_hamiltonian
+
+def _check_tdependent_hamiltonian(hamiltonian) -> Callable[[float],
+        ndarray]:
+    """Input checking for a time-dependent Hamiltonian. """
+    try:
+        tmp_hamiltonian = np.vectorize(hamiltonian)
+        _check_hamiltonian(tmp_hamiltonian(1.0))
+    except Exception as e:
+        raise AssertionError(
+            "Time dependent Hamiltonian must be vectorizable callable.") \
+                from e
+    return tmp_hamiltonian
+
+def _check_tfielddependent_hamiltonian(hamiltonian) -> Callable[[float,
+    complex], ndarray]:
+    try:
+        tmp_hamiltonian = np.vectorize(hamiltonian)
+        _check_hamiltonian(tmp_hamiltonian(1.0, 1.0+1.0j))
+    except Exception as e:
+        raise AssertionError(
+                "Time and field dependent Hamiltonian must be vectorizable "\
+                        "callable.") from e
+    return tmp_hamiltonian
+
+def _check_dissipator_lists(gammas, lindblad_operators) -> Tuple[List, List]:
+    """Check gammas and lindblad operators are lists of equal length."""
+    if gammas is None:
+        gammas = []
+    if lindblad_operators is None:
+        lindblad_operators = []
+    assert isinstance(gammas, list), \
+        "Argument `gammas` must be a list)]."
+    assert isinstance(lindblad_operators, list), \
+        "Argument `lindblad_operators` must be a list."
+    assert len(gammas) == len(lindblad_operators), \
+        "Lists `gammas` and `lindblad_operators` must have the same length."
+    return gammas, lindblad_operators
+
+def _check_gammas_lindblad_operators(gammas, lindblad_operators) -> Tuple[
+        List[float], List[ndarray]]:
+    """Input check for time-independent gammas and lindblad_operators"""
+    # firstly check both are lists of the same length
+    gammas, lindblad_operators = _check_dissipator_lists(gammas,
+            lindblad_operators)
+    try:
+        tmp_gammas = []
+        for gamma in gammas:
+            tmp_gammas.append(float(gamma))
+    except Exception as e:
+        raise AssertionError("All elements of `gammas` must be floats.") \
+            from e
+    try:
+        tmp_lindblad_operators = []
+        for lindblad_operator in lindblad_operators:
+            tmp_lindblad_operators.append(
+                np.array(lindblad_operator, dtype=NpDtype))
+    except Exception as e:
+        raise AssertionError(
+            "All elements of `lindblad_operators` must be numpy arrays.") \
+                from e
+    return tmp_gammas, tmp_lindblad_operators
+
+def _check_tdependent_gammas_lindblad_operators(
+        gammas,
+        lindblad_operators) -> Tuple[List[Callable[[float], float]],
+                List[Callable[[float], ndarray]]]:
+    """Input check for time-dependent gammas and lindblad_operators"""
+    # firstly check both are lists of the same length
+    gammas, lindblad_operators = _check_dissipator_lists(
+            gammas,
+            lindblad_operators)
+    try:
+        tmp_gammas = []
+        for gamma in gammas:
+            float(gamma(1.0))
+            tmp_gamma = np.vectorize(gamma)
+            tmp_gammas.append(tmp_gamma)
+    except Exception as e:
+        raise AssertionError(
+            "All elements of `gammas` must be vectorizable " \
+             + "callables returning floats.") from e
+    try:
+        tmp_lindblad_operators = []
+        for lindblad_operator in lindblad_operators:
+            tmp_lindblad_operator = np.vectorize(lindblad_operator)
+            np.array(tmp_lindblad_operator(1.0))
+            tmp_lindblad_operators.append(tmp_lindblad_operator)
+    except Exception as e:
+        raise AssertionError(
+            "All elements of `lindblad_operators` must be vectorizable " \
+            + "callables returning numpy arrays.") from e
+    return tmp_gammas, tmp_lindblad_operators
+
+def _check_field_eom(dim, field_eom):
+    """Input check a field equation of motion"""
+    test_matrix = _create_density_matrix(dim)
+    test_field = 1.0+1.0j
+    test_time = 1.0
+    try:
+        value = field_eom(test_time, test_matrix, test_field)
+        complex(value)
+    except Exception as e:
+        raise AssertionError("Field equation of motion must "\
+                "take a time, (dim,dim) matrix and field value "\
+                "and return a (complex) scalar.") from e
+    return field_eom
+
+def _liouvillian(hamiltonian, gammas, lindblad_operators):
+    """Lindbladian for a specific Hamiltonian, gammas and lindblad_operators.
+    """
+    liouvillian = -1j * opr.commutator(hamiltonian)
+    for gamma, op in zip(gammas, lindblad_operators):
+        op_dagger = op.conjugate().T
+        liouvillian += gamma * (opr.left_right_super(op, op_dagger) \
+                                - 0.5 * opr.acommutator(np.dot(op_dagger, op)))
+    return liouvillian
+
+def _create_density_matrix(dim, seed=1):
+    r"""Create a repeatable (dim,dim) matrix that represents
+    a valid density matrix :math:`rho`"""
+    rng = np.random.default_rng(seed)
+    a = rng.random((dim, dim)) + 1j*rng.random((dim,dim))
+    b = np.matmul(a, a.conj().T)
+    rho = b / b.trace()
+    return rho
